#include <stdio.h>
#include <stdlib.h>
#include <phidget22.h>
#include "PhidgetHelperFunctions.h"


/**
* Displays info about the attached phidget channel.
* Fired when a Phidget channel with onAttachHandler registered attaches
*
* @param ph The Phidget channel that fired the attach event
* @param *ctx Context pointer
*/
static void CCONV onAttachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	int32_t serialNumber;
	PhidgetHandle hub;
	int32_t hubPort;
	int32_t channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\n[Attach Event]:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nAttach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Displays info about the detached phidget channel.
* Fired when a Phidget channel with onDetachHandler registered detaches
*
* @param ph The Phidget channel that fired the detach event
* @param *ctx Context pointer
*/
static void CCONV onDetachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	PhidgetHandle hub;
	int serialNumber;
	int hubPort;
	int channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\nDetach Event:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Writes phidget error info to stderr.
* Fired when a Phidget channel with onErrorHandler registered encounters an error in the library
*
* @param ph The Phidget channel that fired the error event
* @param *ctx Context pointer
* @param errorCode the code associated with the error of enum type Phidget_ErrorEventCode
* @param *errorString string containing the description of the error fired
*/
static void CCONV onErrorHandler(PhidgetHandle phid, void *ctx, Phidget_ErrorEventCode errorCode, const char *errorString) {

	fprintf(stderr, "[Phidget Error Event] -> %s (%d)\n", errorString, errorCode);
}

/**
* Sets the event handlers for Phidget Attach, Phidget Detach, Phidget Error events
*
* @param ph The Phidget channel to add event handlers to
* @return 0 if the operation succeeds, 1 if it fails
*/
int SetAttachDetachError_Handlers(PhidgetHandle ch) {
	PhidgetReturnCode prc;
	printf("\n--------------------------------------\n");
	printf("\nSetting OnAttachHandler...\n");
	prc = Phidget_setOnAttachHandler(ch, onAttachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Attach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnDetachHandler...\n");
	prc = Phidget_setOnDetachHandler(ch, onDetachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Detach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnErrorHandler...\n");
	prc = Phidget_setOnErrorHandler(ch, onErrorHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Error Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	return 0;
}

/**
* Creates a new instance of a DigitalOutput channel.
*
* @param pdoh Pointer to the PhidgetDigitalOutputHandle channel to create
* @return 0 if the operation succeeds, 1 if it fails
*/
static int CreateDigitalOutput(PhidgetDigitalOutputHandle *pdoh) {
	PhidgetReturnCode prc;

	printf("Creating DigitalOutput Channel...\n");
	prc = PhidgetDigitalOutput_create(pdoh);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Creating DigitalOutput: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}

/**
* Sets the State of the DigitalOutput
*
* @param pdoh The PhidgetDigitalOutputHandle channel to set the State of
* @param state The state to set on the DigitalOutput channel
* @return 0 if the operation succeeds, 1 if it fails
*/
static int SetDigitalOutputState(PhidgetDigitalOutputHandle pdoh, int state) {
	PhidgetReturnCode prc;

	printf("Setting DigitalOutput State to %d\n", state);
	prc = PhidgetDigitalOutput_setState(pdoh, state);

	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Setting DigitalOutput State: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}

/**
* Creates, configures, and opens a DigitalOutput channel.
* Provides interface for setting the State of the DigitalOutput channel
* Closes out DigitalOutput channel
*
* @return 0 if the program exits successfully, 1 if it exits with errors.
*/
int main() {
	PhidgetDigitalOutputHandle ch = NULL;

	/*
	 * Allocate a new Phidget Channel object
	 */

	if (CreateDigitalOutput(&ch))
		goto error;

	/*
	 * Set matching parameters to specify which channel to open
	 */

	if (SetSerialNumber((PhidgetHandle)ch))
		goto error;

	if (SetVINTProperties((PhidgetHandle)ch))
		goto error;

	if (SetChannel((PhidgetHandle)ch))
		goto error;

	if (SetupNetwork((PhidgetHandle)ch))
		goto error;

	/*
	 * Add event handlers before calling open so that no events are missed.
	 */

	if (SetAttachDetachError_Handlers((PhidgetHandle)ch))
		goto error;

	/*
	 * Open the channel with a timeout
	 */

	if (OpenPhidgetChannel_waitForAttach((PhidgetHandle)ch, 5000))
		goto error;

	char buf[100];
	int end = 0;
	int state = 0;

	printf("--------------------\n"
		"\n  | The output of a DigitalOutput channel can be controlled by setting its State.\n"
		"  | The State can be an integer number of either 1 or 0.\n"
		"  | A state of 1 can indicate a high or low output depending on the device used.\n"

		"\nInput a desired state of 0 or 1 and press ENTER\n"
		"Input Q and press ENTER to quit\n");

	while (!end) {
		if (fgets(buf, 100, stdin) == NULL)
			continue;

		switch (buf[0])
		{
		case 'Q':
		case 'q':
			end = 1;
			break;
		case '0':
			state = 0;
			break;
		case '1':
			state = 1;
			break;
		default:
			break;
		}

		if (SetDigitalOutputState(ch, state))
			goto error;
	}

	/*
	 * Perform clean up and exit
	 */
	printf("\nCleaning up...\n");
	ClosePhidgetChannel((PhidgetHandle)ch);
	PhidgetDigitalOutput_delete(&ch);
	printf("\nExiting...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 0;

error:
	PhidgetDigitalOutput_delete(&ch);
	fprintf(stderr, "\nExiting with error(s)...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 1;
}
